import os.path, socket

umpire_root_prefix = os.getenv('UMPIRE_ROOT_DIR', os.path.expanduser("~"))
default_umpire_root = os.path.join(umpire_root_prefix, ".umpire")
default_host_id = socket.gethostname()


# Enable on build nodes, or anywhere that doesn't need SUDO
autoupdate = False

#Cache config
CONFIG_FILENAME = ".umpire"
CONFIG_REPO_SECTION_NAME = "umpire"
CONFIG_ENTRY_SECTION_NAME = "entry"
LOCK_FILENAME = ".umplock"
CURRENT_ENTRY_CONFIG_VERSION = "0.1"
CURRENT_REPO_CONFIG_VERSION = "0.1"
CONFIG_FILENAME = ".umpire"
REMOTE_VERSION_FILENAME = os.path.join(default_umpire_root,"remote_version")
REMOTE_VERSION_URL = "http://s3.amazonaws.com/umpire/version"
LOCKFILE_TIMEOUT = 600 # 10 minutes



import urllib.request, urllib.error, urllib.parse
import sys

from .umpire import __version__
from . import config
import umpire, os
from maestro.core import module
import logging

HELP_KEYS = ["h", "help"]
UPDATE_KEYS = ["u", "update"]
logger = logging.getLogger(__name__)

def parse_version_string(version):
    """
    Returns a tuple containing the major, minor, revision integers
    """
    nums = version.split(".")
    return int(nums[0]), int(nums[1]), int(nums[2])

def __run_pip__():
    from subprocess import call
    call(["pip","install","umpire", "--upgrade"])

def __restart_umpire__():
    from subprocess import call
    args = list(sys.argv)
    args.insert(1, "--skip-update")
    call(args)

class UpdateModule(module.AsyncModule):
    # Required ID of this module
    id = "update"

    def help(self):
        print(self.help_text)
        exit(0)

    def write_remote_version(self, version):
        with open(os.path.join(umpire.get_umpire_root(),"remote_version"), "w+") as f:
            f.write(version)

    def get_remote_version(self):
        return urllib.request.urlopen(config.REMOTE_VERSION_URL).read()

    def run(self,kwargs):
        logger.debug("Running update")
        version = self.get_remote_version()
        self.write_remote_version(version)
        for key in UPDATE_KEYS:
            if key in list(kwargs.keys()):
                __run_pip__()
                __restart_umpire__()



from maestro.core import execute
from maestro.aws import s3
from maestro.tools import path
import logging
import logging.handlers
import logging.config
import argparse

__version__ = "0.8.1"

import sys,os

#Internal Modules
from . import fetch, unpack, deploy, update

from .deploy import HELPTEXT

#Config
from . import config

#Entry point for Umpire
def entry():
    exit_code = Umpire().exit_code
    sys.exit(exit_code)

class Umpire(execute.ModuleExecuter):

    #Defaults for command line flags
    skip_update = True
    debug = False
    deployment_file = None

    #
    # !!!Register all the dependencies of this program here.!!!
    #
    def register_dependencies(self):
        self.register(s3.AsyncS3Downloader())
        self.register(fetch.FetchModule())
        self.register(unpack.UnpackModule())
        self.register(deploy.DeploymentModule())
        self.register(update.UpdateModule())

    def run(self, kwargs):

        sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

        parser = argparse.ArgumentParser(description='Umpire reads a properly formatted JSON deployment file to deploy files.')
        parser.add_argument('-c', '--clear-cache', help='Clear umpire cache', dest='clear', action='store_true', required=False)
        parser.add_argument('-r', '--repair-cache', help='Removes lock files from the cache', dest='repair', action='store_true', required=False)
        parser.add_argument('-v', '--version', help='Displays the current version of Umpire', dest='version',  action='store_true', required=False)
        parser.add_argument('UMPIRE_FILE', help='Umpire json manifest file', nargs='?')
        parser.add_argument('-d', '--debug', help='Run without creating keys or updating keys', dest='debug', action='store_true',
                            required=False)
        args = parser.parse_args()


        log_level = logging.INFO
        if args.debug:
            print('DEBUG logging requested')
            log_level = logging.DEBUG
            self.debug = True
        logger = logging
        FORMAT = '%(asctime)s - %(levelname)s - %(message)s'
        logger.basicConfig(format=FORMAT, stream=sys.stdout, level=log_level)


        if args.clear:
            import shutil
            logger.info("Clearing cache...")
            shutil.rmtree(get_umpire_root())
            logger.info("Completed")
            sys.exit(0)

        if args.repair:
            logger.info("Repairing cache...")
            path.purge(config.LOCK_FILENAME, get_umpire_root())
            logger.info("Completed")
            sys.exit(0)

        if args.version:
            from subprocess import call
            call(["pip3","show","umpire"])
            sys.exit(0)
        # elif item == "-s" or item == "--skip-update":
        #     self.skip_update = True


        if args.UMPIRE_FILE:
            self.deployment_file = args.UMPIRE_FILE
        else:
            parser.print_help()
            sys.exit(0)
        logger.info("Beginning Umpire deploy")
        self.register_dependencies()
        if not self.skip_update:
            updater = self.getObject("update")
            try:
                #Check if there's an update available
                with open(os.path.join(config.REMOTE_VERSION_FILENAME), "r") as f:
                    version = f.read()
                    rmajor, rminor, rrev = update.parse_version_string(version)
                    major, minor, rev = update.parse_version_string(__version__)
                    logger.info (str(rmajor) + str(rminor) + str(rrev))
                    logger.info (str(major) + str(minor) + str(rev))
                    if (rmajor > major) or ((rminor > minor) and (rmajor == rmajor)) or ((rrev > rev) and (rminor == minor) and (rmajor == rmajor)):
                        if config.autoupdate:
                            logger.info("\n!!!Update (%s) detected!!!\n" % version[:-1])
                            logger.info("Autoupdating...\n")
                            updater.run({"update":None})
                            sys.exit(0)
                        else:
                            logger.info("!!!Update (%s) detected!!!" % version[:-1])
                            logger.info("Please run pip install umpire --upgrade as an Administrator.\n")
                            updater.start()
            except IOError:
                updater.start()



        #Get the instance of our first module
        deployer = self.getObject("deploy")
        deployer.cache_root = get_umpire_root()
        deployer.deployment_file = self.deployment_file

        deployer.log_level = log_level
        self.exit_code = deployer.run(kwargs)
        logger.info("Completed")

def get_umpire_root():
    hardcoded_root = "./umpire_tmp"
    try:
        if not os.path.exists(config.default_umpire_root):
            os.makedirs(config.default_umpire_root)
        return config.default_umpire_root
    except Exception as e:
        logging.error("Umpire: Error obtaining default umpire root location, using local directory './umpire_tmp': " + str(e), file=sys.stderr)
        if not os.path.exists(hardcoded_root):
            os.makedirs(hardcoded_root)
        return hardcoded_root

if __name__ == '__main__':
    entry()


"""fetch.py"""

import sys, os, urllib, time, traceback
from . import cache, config
from multiprocessing import Value
from maestro.core import module
from maestro.aws import s3
from maestro.tools import file as mfile
import logging
logger = logging.getLogger(__name__)

CACHE_LOCATION_KEYS = ["c", "cache-location"]
DEPENDENCY_NAME_KEYS = ["n", "name"]
DEPENDENCY_PLATFORM_KEYS = ["p", "platform"]
DEPENDENCY_REPOSITORY_KEYS = ["r", "repository"]
DEPENDENCY_VERSION_KEYS = ["v", "version"]
HELP_KEYS = ["h", "help"]

class EntryState(object):
    CACHE = 0
    DOWNLOADED = 1
    UPDATED = 2

class EntryError(Exception):
    pass

class FetchModule(module.AsyncModule):

    # Required ID of this module
    id = "fetch"

    #Initialize shm value #TODO: Near future
    progress = Value('d', 0.0)

    # Version of dependency (Default latest) #TODO: Future
    dependency_version = None

    ### Required Variables for this Module ###
    #Location of the cache root folder (r+w by every user)
    cache_root = None

    #Name of dependency to fetch
    dependency_name = None

    #Name of platform of dependency to fetch
    dependency_platform = None

    #The repository to fetch the dependency, or to verify the integrity from
    dependency_repo = None

    #Is the dependency going to be linked?
    dependency_is_link = None

    #Is the dependency going to be extracted after the download
    dependency_unpack = None

    #Are we going to check if this dependency has changed in the remote?
    keep_updated = None

    log_level=None

    def run(self,kwargs):
        logger = logging
        FORMAT = '%(asctime)s - %(levelname)s - %(message)s'
        logger.basicConfig(format=FORMAT, stream=sys.stdout, level=self.log_level)
        logger.debug("Running Fetch")
        #Verify argument validity
        self.__verify_arguments__()

        #Get cache name from remote url
        cache_name = self.get_cache_name()

        #Get cache path
        cache_path = os.path.join(self.cache_root, cache_name)

        logger.debug(f"Cache path: {cache_path}")

        #Get cache object (will raise an exception if it doesn't exist)
        cache_obj = cache.LocalCache(cache_path, host_id=config.default_host_id, log_level=self.log_level)

        full_url = s3.join_s3_url(self.dependency_repo, self.dependency_platform, self.dependency_name, self.dependency_version)

        cache_obj.lock(os.path.join(cache_path,self.dependency_platform, self.dependency_name, self.dependency_version))
        #Try to get entry from cache
        entry = cache_obj.get(self.dependency_platform, self.dependency_name, self.dependency_version)

        #Set state
        if entry is None:
            state = EntryState.DOWNLOADED
        else:
            state = EntryState.CACHE

        #Verify Remote sha256 if updating
        if self.keep_updated and state == EntryState.CACHE:
            bucket,prefix = s3.parse_s3_url(full_url)
            checksums = list()
            try:
                for file in s3.find_files(bucket, prefix, anonymous=False, sha256=True ):
                    checksums.append(file[1])
            except:
                logger.debug(f"{traceback.print_exc()}")
                for file in s3.find_files(bucket, prefix, anonymous=True, sha256=True ):
                    checksums.append(file[1])

            if entry.md5 not in checksums:
                state = EntryState.UPDATED

        #We need to download the file, it wasn't in the cache
        if state == EntryState.DOWNLOADED or state == EntryState.UPDATED:

            if state == EntryState.DOWNLOADED:
                logger.info(self.format_entry_name() + ": Not in cache")
            else:
                logger.info(self.format_entry_name() + ": Cache is obsolete")

            logger.info(self.format_entry_name() + ": Downloading " + full_url)

            #Get Downloader
            downloader = s3.AsyncS3Downloader(None)

            #Set Downloader arguments
            downloader.source_url = full_url+"/"
            downloader.sha256 = True
            downloader.destination_path = os.path.join(self.cache_root, "downloading") + os.sep
            downloader.start()

            #Wait for downloader to finish #TODO: Do something with the reported progress
            while downloader.status != 2:
                time.sleep(0.5)

            #Check for an exception, if so bubble it up
            if downloader.exception is not None:
                logger.error(f"Exception!!! {downloader.exception}")
                # raise downloader.exception
                sys.exit(1)
            logger.info(self.format_entry_name() + ": Download complete")
            logger.debug(f"Result: {downloader.result}")
            if downloader.result is None or len(downloader.result) == 0:
                raise EntryError(self.format_entry_name() + ": Unable to find remote entry '" + full_url + "'")

            #Iterate of the result (downloaded files)
            for item, checksum in downloader.result:
                local_file_checksum = mfile.sha256_checksum(item, checksum)
                logging.debug(f"sha 256 checksum local_file: {local_file_checksum} remote: {checksum}")
                if checksum != local_file_checksum:
                    logger.warning(self.format_entry_name() + ": WARNING: Downloaded file does not match the checksum on the server")
                    logger.warning(self.format_entry_name() + ": WARNING: local:\t" + str(local_file_checksum))
                    logger.warning(self.format_entry_name() + ": WARNING: server:\t" + str(checksum))
                if self.dependency_unpack:
                    print(self.format_entry_name() + ": Unpacking...")
                    logger.info (self.format_entry_name() + ": Unpacking...")
                #Put will unlock
                cache_obj.put(item,self.dependency_platform, self.dependency_name, self.dependency_version, unpack_bol=self.dependency_unpack, checksum=checksum)
            entry = cache_obj.get(self.dependency_platform, self.dependency_name, self.dependency_version)
            if entry is None:
                raise EntryError(self.format_entry_name() + ": Error retrieving entry from cache.")
        cache_obj.unlock(os.path.join(cache_path,self.dependency_platform, self.dependency_name, self.dependency_version))
        #Entry is not None, return all the files listed in the entry that aren't the configuration files
        return [os.path.abspath(os.path.join(entry.path,f)) for f in os.listdir(entry.path) if f != ".umpire"], state


    def format_entry_name(self):
        return str(self.dependency_platform) + "/" + str(self.dependency_name) + " " + str(self.dependency_version)

    def __verify_arguments__(self):
        if not self.dependency_repo:
            raise ValueError("You must specify a valid repository URL.")
        if not self.dependency_name:
            raise ValueError("You must specify a valid name for the dependency.")
        if not self.dependency_platform:
            raise ValueError("You must specify a valid platform for the dependency.")
        if not self.dependency_version:
            raise ValueError("You must specify a valid version for the dependency.")
        if not self.dependency_is_link:
            # Not used
            pass
        if not self.keep_updated:
            self.keep_updated = False
        if not isinstance(self.dependency_unpack,bool):
            logger.info(str(self.dependency_unpack))
            raise ValueError("You must specify whether the dependency should be unpacked or not")

    def get_cache_name(self):
        try:
            bucket, prefix = s3.parse_s3_url(self.dependency_repo)
            return bucket + ".s3"
        except ValueError:
            raise ValueError("Only URLs formatted as: s3://{bucket}/ are currently accepted.")

    def __parse_kwargs__(self,kwargs):
        if kwargs is None:
            return
        for key, val in kwargs.items():
            if key in CACHE_LOCATION_KEYS:
                self.cache_repo = val
            elif key in DEPENDENCY_NAME_KEYS:
                self.dependency_name = val
            elif key in DEPENDENCY_PLATFORM_KEYS:
                self.dependency_platform = val
            elif key in DEPENDENCY_REPOSITORY_KEYS:
                self.dependency_repo = val
            elif key in DEPENDENCY_VERSION_KEYS:
                self.dependency_version = val




"""unpack.py"""

HELPTEXT = """
     ----- Unpack Module -----

The unpack module unpacks a tar gz or zip file.

Usage: Currently N/A

"""

import sys, os, subprocess
from maestro.core import module
from tqdm import tqdm
import logging, tarfile, zipfile

HELP_KEYS = ["h", "help"]

class UnpackError(Exception):
    pass

class UnpackModule(module.AsyncModule):
    # Required ID of this module
    id = "unpack"
    file_path = None
    destination_path = None
    delete_archive = False
    log_level = None

    def help(self):
        print(self.help_text)
        exit(0)

    def untar_gzip(self):
        with tarfile.open(name=self.file_path) as tar:
            for member in tqdm(iterable=tar.getmembers(), desc=f'Extracting {self.file_path}', total=len(tar.getmembers())):
                tar.extract(member=member, path=self.destination_path)

    def unzip(self):
        with zipfile.ZipFile(self.file_path) as zf:
            for member in tqdm(zf.infolist(), desc=f'Extracting {self.file_path}'):
                try:
                    zf.extract(member, self.destination_path)
                except zipfile.error as e:
                    pass

    def run(self,kwargs):
        logger = logging
        FORMAT = '%(asctime)s - %(levelname)s - %(message)s'
        logger.basicConfig(format=FORMAT, stream=sys.stdout, level=self.log_level)
        logger.debug("Running Unpack")
        if self.file_path.endswith("tar.gz") or self.file_path.endswith("tgz"):
            self.untar_gzip()
        elif self.file_path.endswith(".zip"):
            logger.debug("Untar")
            self.unzip()
        else:
            raise UnpackError("Unable to unpack this type of file: " + os.path.split(self.file_path)[1])

        if self.delete_archive is True:
            os.remove(self.file_path)
